/*****************************\
| permutationTest.cpp         |
|                             |
| AUTHOR: Byron Kuo           |
|                             |
| Inimex Pharmaceuticals Inc. |
\*****************************/

// This permutation test program takes in the data from two groups
// and performs permutation test with the specified number of iterations.
// The number of iterations is used to generate the distribution, which is
// assumed normal depending on the quality of the random number generator.
//
// For each iteration, the datasets (input 1 + input 2) are resampled without
// replacement to simulate the two datasets.  The difference of the average
// simulated means is computed and stored.  The p-value is computed as the
// percentage of the simulated means which are greater than the actual observed
// means relative to the number of simulated samples.
//
// The result is written to a file with the gene names and the p-values.
//
// INPUT Parameters:
//   [gene list]: One column.  Each line consists of the name of each gene.
//                This list can be produced by stripping off the first column
//                of the ArrayVision output file and storing the column
//                into a separate file.
//
//   [input file 1]: Either the control or the test dataset.  Each column
//                   represents a sample.  Tab-delimited.
//
//   [input file 1 size]: The number of columns(samples) in the first
//                        input file.
//
//   [input file 2]: Either the control or the test dataset.  Each column
//                   represents a sample.  Tab-delimited.
//
//   [input file 2 size]: The number of columns(samples) in the second
//                        input file.
//
//   [output file]: The output file name.
//
//   [number of iterations]: The number of samples for each gene the user
//                           wishes to generate the distribution for p-value
//                           computation.
//
// OUTPUT:
//   With the specified file name, the output file consists of each gene
//   name and the p-value calculated from the permutations.
//
// ASSUMPTIONS:
//   It is assumed that in [gene list], [input file 1], and [input file 2],
//   the number of lines are equal and the order of the genes are kept.
//

#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <string.h>
#include <list>
#include <cstdlib>
#include <time.h>
#include <math.h>
#include <string>
using namespace std;

#define confidence 5

// FUNCTION: compute p-value from the given sorted array and the given observed value
// INPUT: 'array' is the array of the simulated differences, sorted in ascending order
//        'size' is the size of the array
//        'observed' is the actual observed difference
// OUTPUT: the p-value of float type
float computePvalue (float* array, int size, float observed)
{
  int count = 0;

  // count the number of simulated differences which are smaller than the observed
  while (count < size) {
    if (array[count] > observed) {
      break;
    }
    count++;
  }

  // p-value is the percentage of the simulated differences that are greater than
  // the observed, relative to the total number of simulated differences
  float p_value = (float(size)-float(count))/float(size);

  return p_value;
}

// FUNCTION: comparing function used for quick sort
int compare (const void * a, const void * b)
{
  return ( *(int*)a - *(int*)b );
}

// FUNCTION: perform permutation test for one gene
// INPUT: 'array' is an array of pooled datasets of one gene
//        'size1' is the number of samples for the first dataset
//        'size2' is the number of samples for the second dataset
//        'numIteration' is the number of resamples
// OUTPUT: the p-value of float type
float permutationTest (float* array, int size1, int size2, int numIteration)
{
  int totalSize = size1 + size2;
  float* distributionArray = new float[numIteration]; // storing the final distribution
  int distributionIndex = 0;
  float p_value;
  float sum1 = 0;
  float sum2 = 0;
  float sumAll;
  float mean1;
  float mean2;
  float obsDiff; // observed difference

  // compute mean1
  for (int i = 0; i < size1; i++) {
    sum1 = sum1 + array[i];
  }
  mean1 = sum1/size1;

  // compute mean2
  for (int i = size1; i < totalSize; i++) {
    sum2 = sum2 + array[i];
  }
  mean2 = sum2/size2;
  obsDiff = fabsf(mean1 - mean2);
  sumAll = sum1 + sum2;

  // if size1 > size2, swap to increase efficiency
  if (size1 > size2) {
    int temp = size1;
    size1 = size2;
    size2 = temp;
  }

  // perform resampling
  for (int i = 0; i < numIteration; i++) {
    float* permutationArray = new float[totalSize]; // used for each permutation
    int* usedArray = new int[100]; // used to temporarily storing array indices
                                   // that have already been picked
    int usedSize = 0; // size of teh usedArray
    float simulateSum1 = 0;
    float simulateSum2 = 0;
    float simulateMean1;
    float simulateMean2;
    float simulateDiff;
    float simulateDiffSum = 0;

    while (usedSize < size1) {
      int indexRandom = rand()%totalSize;
      bool checkExist = false;

      // check if the random index has been used before
      for (int j = 0; j < usedSize; j++) {

	if (indexRandom == usedArray[j]) {
	    checkExist = true;
	    break;
	}
      }

      // if the random index has not been used, then sum the value
      // and add the index to usedArray[]
      if (checkExist == false) {

	usedArray[usedSize] = indexRandom;
	simulateSum1 += array[indexRandom];
	usedSize++;
      }
    }

    simulateSum2 = sumAll - simulateSum1;

    // compute means and difference
    simulateMean1 = simulateSum1 / size1;
    simulateMean2 = simulateSum2 / size2;
    simulateDiff = fabsf(simulateMean1 - simulateMean2);
    simulateDiffSum += simulateDiff;

    // store the simulated difference into the distribution array
    distributionArray[distributionIndex] = simulateDiff;
    distributionIndex++;

    // clean up garbages
    delete permutationArray;
    delete usedArray;
  }
  // sort the distribution array
  qsort(distributionArray, distributionIndex, sizeof(float), compare);

  p_value = computePvalue(distributionArray, distributionIndex, obsDiff);

  // clean up garbage
  delete distributionArray;

  return p_value;
}


// MAIN FUNCTION
int main(int argc, char *argv[])
{

  // check for correct number of arguments
  if (argc != 8) {
    cout << "Incorrect number of arguments\n";
    cout << "Usage: [command] [gene list] [input file 1] [input 1 size] [input file 2] [input 2 size] [output file] [number of iterations]\n";
    return 0;
  }

  //  const int seed = time(NULL); // seed for random number
  const int seed = 9999; // seed for random number
  srand(seed);
  int numIteration = atoi(argv[7]); // number of iterations as specified

  // input/output file initiations
  ifstream genelist;
  ifstream infile1;
  ifstream infile2;
  ofstream outfile;

  // open the files
  genelist.open(argv[1]);
  infile1.open(argv[2]);
  infile2.open(argv[4]);
  outfile.open(argv[6]);

  string delimiter = "\t";
  int size1 = atoi(argv[3]);
  int size2 = atoi(argv[5]);
  int totalSize = size1 + size2; // overall array size
  float *combineArray; // array declaration


  string line1;
  string line2;
  while(getline(infile1, line1)) {

    //-------------------- go through first file -----------------------//
    string str;
    list<float> L1;

    int startPos = 0;
    int pos = line1.find(delimiter);
    for (int i = 0; i < size1; i++) {
      str = line1.substr(startPos, pos-startPos);
      L1.push_back (atof(str.c_str()));
      startPos = pos + delimiter.length();
      pos = line1.find(delimiter, startPos);
    }

    
    //------------------- go through second file -------------------------//
    getline(infile2, line2);
    startPos = 0;
    pos = line2.find(delimiter, startPos);

    list<float> L2;
    startPos = 0;
    pos = line2.find(delimiter);
    for (int i = 0; i < size2; i++) {
      str = line2.substr(startPos, pos-startPos);
      L2.push_back (atof(str.c_str()));
      startPos = pos + delimiter.length();
      pos = line2.find(delimiter, startPos);
    }

    combineArray = new float[totalSize];

    int tempSize = 0;
    list<float>::iterator i;
    for(i=L1.begin(); i != L1.end(); i++){
      combineArray[tempSize] = *i;
      tempSize++;
    }
    for (i=L2.begin(); i != L2.end(); i++){
      combineArray[tempSize] = *i;
      tempSize++;
    }
    /*
    for (int i = 0; i < totalSize; i++) {
      cout << combineArray[i] << " ";
    }
    */
    //cout << endl;
    


  float p_value = permutationTest(combineArray, size1, size2, numIteration);

  string gene;
  getline(genelist, gene);
  outfile << gene << "\t" << p_value << endl;
  //  cout << p_value << endl;
  }


  // close the files
  genelist.close();
  infile1.close();
  infile2.close();
  outfile.close();
  return 0;
}
